from time import sleep


def test_snippets(browser):
    browser.visit('/snippets')
    browser.wait_for_js_variable('initFormSnippets')
    browser.execute_script("""
        initFormSnippets(document.querySelector('.formcode-snippets'));
    """)

    assert browser.is_element_present_by_css('.formcode-snippets')
    assert len(browser.find_by_css('.formcode-toolbar')) == 1

    assert not browser.find_by_css('.formcode-snippet')
    browser.find_by_css('.formcode-toolbar-element').click()

    assert browser.find_by_css('.formcode-snippet')
    browser.find_by_css('.formcode-toolbar-element').click()

    assert not browser.find_by_css('.formcode-snippet')
    browser.find_by_css('.formcode-toolbar-element').click()

    browser.find_by_css('.formcode-snippet-optional').click()
    assert '= ___' in browser.find_by_css('textarea').value

    assert not browser.find_by_css('.formcode-snippet')
    browser.find_by_css('.formcode-toolbar-element').click()

    browser.find_by_css('.formcode-snippet-required').click()
    assert '*= ___' in browser.find_by_css('textarea').value


def test_registry(browser):
    browser.visit('/registry')
    browser.wait_for_js_variable('formcodeWatcherRegistry')
    browser.execute_script("""
        var watcher = formcodeWatcherRegistry.new("test");
        var unsubscribe = watcher.subscribe(function(value) {
            window.code = value;
        });
        watcher.update("Label = ...");
        unsubscribe();
        watcher.update("Label = ___");
    """)

    browser.wait_for_js_variable('window.code')
    code = browser.evaluate_script("window.code")
    assert code == [{'human_id': 'Label', 'type': 'textarea', 'id': 'label'}]


def test_formcode_format(browser):
    browser.visit('/formcode-format')
    browser.wait_for_js_variable('initFormcodeFormat')
    browser.execute_script("""
        var watcher = formcodeWatcherRegistry.new("test");
        var el = document.querySelector('#container');

        el.setAttribute('data-watcher', 'test');
        el.setAttribute('data-target', 'textarea');

        initFormcodeFormat(el);

        watcher.update('Textfield = ___');
    """)

    browser.find_by_css('.formcode-toolbar-element').click()
    browser.find_by_css('.formcode-snippet').click()

    assert browser.find_by_css('textarea').value == '[Textfield]'


def test_formcode_select_empty(browser):
    browser.visit('/formcode-select')
    browser.wait_for_js_variable('initFormcodeSelect')
    browser.driver.execute_script("""
        var watcher = formcodeWatcherRegistry.new();
        var el = document.querySelector('#container');

        initFormcodeSelect(el, watcher, 'textarea', ['text', 'textarea']);
        watcher.update(arguments[0]);
    """, 'A = ___\nB = ...\nC = *.png')

    assert len(browser.find_by_css('.formcode-select input')) == 2
    browser.find_by_css('.formcode-select input')[0].click()
    browser.find_by_css('.formcode-select input')[1].click()

    assert browser.find_by_css('textarea').value == "A\nB"

    browser.find_by_css('.formcode-select input')[1].click()
    assert browser.find_by_css('textarea').value == "A"

    browser.find_by_css('.formcode-select input')[1].click()
    assert browser.find_by_css('textarea').value == "A\nB"

    browser.find_by_css('.formcode-select input')[0].click()
    assert browser.find_by_css('textarea').value == "B"

    browser.find_by_css('.formcode-select input')[1].click()
    assert browser.find_by_css('textarea').value == ""


def test_formcode_select_prefilled(browser):
    browser.visit('/formcode-select')
    browser.wait_for_js_variable('initFormcodeSelect')
    browser.driver.execute_script("""
        var watcher = formcodeWatcherRegistry.new();
        var el = document.querySelector('#container');
        document.querySelector('textarea').value='A'

        initFormcodeSelect(el, watcher, 'textarea', ['text', 'textarea']);
        watcher.update(arguments[0]);
    """, 'A = ___\nB = ...\nC = *.png')

    assert len(browser.find_by_css('.formcode-select input:checked')) == 1


def test_formcode_keep_selection(browser):
    browser.visit('/formcode-select')
    browser.wait_for_js_variable('initFormcodeSelect')
    browser.driver.execute_script("""
        var watcher = document.watcher = formcodeWatcherRegistry.new();
        var el = document.querySelector('#container');
        document.querySelector('textarea').value='A'

        initFormcodeSelect(el, watcher, 'textarea', ['text', 'textarea']);
        watcher.update('B = ___');
    """)

    assert len(browser.find_by_css('.formcode-select input:checked')) == 0
    browser.driver.execute_script("document.watcher.update('A = ___');")

    sleep(0.1)

    assert len(browser.find_by_css('.formcode-select input:checked')) == 1
    browser.driver.execute_script("document.watcher.update('C = ___');")

    sleep(0.1)

    assert len(browser.find_by_css('.formcode-select input:checked')) == 0


def test_field_errors_should_not_yield_updates(browser):
    browser.visit('/formcode-format')
    browser.wait_for_js_variable('initFormcodeFormat')
    browser.execute_script("""
        document.watcher = formcodeWatcherRegistry.new();
        var el = document.querySelector('#container')

        initFormcodeFormat(el, document.watcher, 'textarea');
        document.watcher.update('Textfield = ___');
    """)

    browser.find_by_css('.formcode-toolbar-element').click()

    assert len(browser.find_by_css('.formcode-snippet')) == 1
    assert browser.find_by_css('.formcode-snippet').text == "Textfield"

    browser.execute_script("document.watcher.update('Test =-= !invalid');")

    assert len(browser.find_by_css('.formcode-snippet')) == 1
    assert browser.find_by_css('.formcode-snippet').text == "Textfield"

    browser.execute_script("document.watcher.update('Fixed = ___');")

    assert len(browser.find_by_css('.formcode-snippet')) == 1
    assert browser.find_by_css('.formcode-snippet').text == "Fixed"



class FormError(Exception):
    pass


class DuplicateLabelError(FormError):

    def __init__(self, label):
        self.label = label

    def __repr__(self):
        return "DuplicateLabelError(label='{}')".format(self.label)


class InvalidMimeType(FormError):
    pass


class UnableToComplete(FormError):
    pass


class InvalidFormSyntax(FormError):
    def __init__(self, line):
        self.line = line



import sedate

from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.form.models.submission import FormSubmission
from sqlalchemy import and_
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import or_
from sqlalchemy import Text
from sqlalchemy import text
from sqlalchemy.dialects.postgresql import ExcludeConstraint
from sqlalchemy.orm import object_session, relationship
from sqlalchemy.schema import CheckConstraint
from sqlalchemy.sql.elements import quoted_name
from uuid import uuid4


daterange = Column(quoted_name('DATERANGE("start", "end")', quote=False))


class FormRegistrationWindow(Base, TimestampMixin):
    """ Defines a registration window during which a form definition
    may be used to create submissions.

    Submissions created thusly are attached to the currently active
    registration window.

    Registration windows may not overlap.

    """

    __tablename__ = 'registration_windows'

    #: the public id of the registraiton window
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the name of the form to which this registration window belongs
    name = Column(Text, ForeignKey("forms.name"), nullable=False)

    #: true if the registration window is enabled
    enabled = Column(Boolean, nullable=False, default=True)

    #: the start date of the window
    start = Column(Date, nullable=False)

    #: the end date of the window
    end = Column(Date, nullable=False)

    #: the timezone of the window
    timezone = Column(Text, nullable=False, default='Europe/Zurich')

    #: the number of spots (None => unlimited)
    limit = Column(Integer, nullable=True)

    #: enable an overflow of submissions
    overflow = Column(Boolean, nullable=False, default=True)

    #: submissions linked to this
    submissions = relationship('FormSubmission', backref='registration_window')

    __table_args__ = (

        # ensures that there are no overlapping date ranges within one form
        ExcludeConstraint(
            (name, '='), (daterange, '&&'),
            name='no_overlapping_registration_windows',
            using='gist'
        ),

        # ensures that there are no adjacent date ranges
        # (end on the same day as next start)
        ExcludeConstraint(
            (name, '='), (daterange, '-|-'),
            name='no_adjacent_registration_windows',
            using='gist'
        ),

        # ensures that start <= end
        CheckConstraint(
            '"start" <= "end"',
            name='start_smaller_than_end'
        ),
    )

    @property
    def localized_start(self):
        return sedate.align_date_to_day(
            sedate.standardize_date(
                sedate.as_datetime(self.start), self.timezone
            ), self.timezone, 'down'
        )

    @property
    def localized_end(self):
        return sedate.align_date_to_day(
            sedate.standardize_date(
                sedate.as_datetime(self.end), self.timezone
            ), self.timezone, 'up'
        )

    def accepts_submissions(self, required_spots=1):
        assert required_spots > 0

        if not self.enabled:
            return False

        if not self.in_the_present:
            return False

        if self.overflow:
            return True

        if self.limit is None:
            return True

        return self.available_spots >= required_spots

    def disassociate(self):
        """ Disassociates all records linked to this window. """

        for submission in self.submissions:
            submission.disclaim()
            submission.registration_window_id = None

    @property
    def available_spots(self):
        return max(self.limit - self.claimed_spots - self.requested_spots, 0)

    @property
    def claimed_spots(self):
        """ Returns the number of actually claimed spots. """

        return object_session(self).execute(text("""
            SELECT SUM(
                COALESCE(claimed, 0)
            )
            FROM submissions
            WHERE registration_window_id = :id
            AND submissions.state = 'complete'
        """), {'id': self.id}).scalar() or 0

    @property
    def requested_spots(self):
        """ Returns the number of requested spots.

        When the claim has not been made yet, `spots` are counted as
        requested. When the claim has been partially made, the difference is
        counted as requested. If the claim has been fully made, the result is
        0. If the claim has been relinquished, the result is 0.

        """
        return object_session(self).execute(text("""
            SELECT GREATEST(
                SUM(
                    CASE WHEN claimed IS NULL THEN spots
                         WHEN claimed = 0 THEN 0
                         ELSE spots - claimed
                    END
                ), 0
            )
            FROM submissions
            WHERE registration_window_id = :id
            AND submissions.state = 'complete'
        """), {'id': self.id}).scalar() or 0

    @property
    def next_submission(self):
        """ Returns the submission next in line. In other words, the next
        submission in order of first come, first serve.

        """

        q = object_session(self).query(FormSubmission)
        q = q.filter(FormSubmission.registration_window_id == self.id)
        q = q.filter(FormSubmission.state == 'complete')
        q = q.filter(or_(
            FormSubmission.claimed == None,
            and_(
                FormSubmission.claimed > 0,
                FormSubmission.claimed < FormSubmission.spots,
            )
        ))
        q = q.order_by(FormSubmission.created)

        return q.first()

    @property
    def in_the_future(self):
        return sedate.utcnow() <= self.localized_start

    @property
    def in_the_past(self):
        return self.localized_end <= sedate.utcnow()

    @property
    def in_the_present(self):
        return self.localized_start <= sedate.utcnow() <= self.localized_end



import pytest
import re

from decimal import Decimal
from onegov.form.utils import decimal_range
from onegov.form.parser.grammar import (
    checkbox,
    code,
    currency,
    date,
    datetime,
    decimal,
    decimal_range_field,
    email,
    field_identifier,
    fileinput,
    integer_range_field,
    password,
    radio,
    stdnum,
    text_without,
    textarea,
    textfield,
    time,
    with_whitespace_inside,
    url
)


def test_text_without():
    assert text_without('?!').parseString('what')[0] == 'what'
    assert text_without('?!').parseString('what what?')[0] == 'what'
    assert text_without('?!').parseString('what!')[0] == 'what'
    assert text_without('?').parseString('what!')[0] == 'what!'


def test_with_whitespace_inside():
    text = text_without('')
    assert with_whitespace_inside(text).parseString("a b")[0] == "a b"
    assert with_whitespace_inside(text).parseString("a b ")[0] == "a b"
    assert with_whitespace_inside(text).parseString("a  b ")[0] == "a"


def test_field_identifier():
    parse = field_identifier().parseString

    assert parse("Yes?=").asDict() == {'required': False, 'label': 'Yes?'}
    assert parse("Yes?*=").asDict() == {'required': True, 'label': 'Yes?'}

    assert parse("OMG. U ok?! =").label == "OMG. U ok?!"
    assert parse("a b =").label == "a b"
    assert parse("ab =").label == "ab"
    assert parse("1 = 2").label == "1"
    assert parse("what=").label == "what"
    assert parse("what=*").label == "what"
    assert parse("what*=").label == "what"
    assert parse("What* =").label == 'What'
    assert parse("Sure?! =").label == 'Sure?!'

    assert not parse("Do you?! =").required
    assert not parse("what=*").required
    assert parse("What* =").required
    assert parse("What* =").required
    assert parse("What* =").required


def test_textfield():

    field = textfield()

    f = field.parseString('___')
    assert f.type == 'text'
    assert not f.length
    assert f.asDict() == {'type': 'text'}

    f = field.parseString('___[25]')
    assert f.type == 'text'
    assert f.length == 25
    assert f.asDict() == {'length': 25, 'type': 'text'}

    # valid regex
    f = field.parseString('___/[0-9]{4}')
    assert f.type == 'text'
    assert not f.length
    assert f.asDict() == {
        'type': 'text',
        'regex': re.compile('[0-9]{4}')
    }

    # invalid regex
    with pytest.raises(re.error):
        f = field.parseString('___/[0-9')

    # combined with a length validator
    f = field.parseString('___[10]/[a-zA-Z]+')
    assert f.type == 'text'
    assert f.length == 10
    assert f.asDict() == {
        'length': 10,
        'type': 'text',
        'regex': re.compile('[a-zA-Z]+')
    }


def test_textarea():

    field = textarea()

    f = field.parseString("...")
    assert f.type == 'textarea'
    assert not f.rows
    f.asDict() == {'type': 'textarea'}

    f = field.parseString("...[15]")
    assert f.type == 'textarea'
    assert f.rows == 15
    f.asDict() == {'rows': 15, 'type': 'textarea'}


def test_password():

    field = password()

    f = field.parseString("***")
    assert f.type == 'password'
    assert f.asDict() == {'type': 'password'}


def test_email():

    field = email()

    f = field.parseString("@@@")
    assert f.type == 'email'
    assert f.asDict() == {'type': 'email'}


def test_url():

    field = url()

    f = field.parseString("http://")
    assert f.type == 'url'
    assert f.asDict() == {'type': 'url'}

    f = field.parseString("https://")
    assert f.type == 'url'
    assert f.asDict() == {'type': 'url'}


def test_dates():

    field = date().searchString('YYYY.MM.DD')
    field.asDict() == {'type': 'date', 'label': 'Date'}

    field = datetime().searchString('YYYY.MM.DD HH:MM')
    field.asDict() == {'type': 'datetime', 'label': 'Datetime'}

    field = time().searchString('HH:MM')
    field.asDict() == {'type': 'time', 'label': 'Time'}


def test_stdnum():
    field = stdnum()

    f = field.parseString("#test")
    assert f.type == 'stdnum'
    assert f.format == 'test'
    assert f.asDict() == {'type': 'stdnum', 'format': 'test'}

    f = field.parseString("# test")
    assert f.type == 'stdnum'
    assert f.format == 'test'
    assert f.asDict() == {'type': 'stdnum', 'format': 'test'}

    f = field.parseString("# asdf.asdf")
    assert f.type == 'stdnum'
    assert f.format == 'asdf.asdf'
    assert f.asDict() == {'type': 'stdnum', 'format': 'asdf.asdf'}


def test_radio():

    field = radio()

    f = field.parseString("( ) Male")
    assert f.type == 'radio'
    assert f.label == 'Male'
    assert not f.checked

    f = field.parseString("(x) Space Alien")
    assert f.type == 'radio'
    assert f.label == 'Space Alien'
    assert f.checked


def test_checkbox():

    field = checkbox()

    f = field.parseString("[x] German")
    assert f.type == 'checkbox'
    assert f.label == 'German'
    assert f.checked

    f = field.parseString("[ ] Swiss German")
    assert f.type == 'checkbox'
    assert f.label == 'Swiss German'
    assert not f.checked


def test_fileinput():

    field = fileinput()

    f = field.parseString("*.*")
    assert f.type == 'fileinput'
    assert f.extensions == ['*']

    f = field.parseString("*.pdf")
    assert f.type == 'fileinput'
    assert f.extensions == ['pdf']

    f = field.parseString("*.bat")
    assert f.type == 'fileinput'
    assert f.extensions == ['bat']

    f = field.parseString("*.png|*.jpg|*.gif")
    assert f.type == 'fileinput'
    assert f.extensions == ['png', 'jpg', 'gif']


def test_prices():
    field = radio()

    f = field.parseString("( ) Default Choice (100 CHF)")
    assert f.type == 'radio'
    assert f.label == 'Default Choice'
    assert not f.checked
    assert f.pricing.amount == Decimal('100.00')
    assert f.pricing.currency == 'CHF'

    f = field.parseString("(x) Luxurious Choice (200 CHF)")
    assert f.type == 'radio'
    assert f.label == 'Luxurious Choice'
    assert f.checked
    assert f.pricing.amount == Decimal('200.00')
    assert f.pricing.currency == 'CHF'

    field = checkbox()

    f = field.parseString("[x] Extra Luggage (150.50 USD)")
    assert f.type == 'checkbox'
    assert f.label == 'Extra Luggage'
    assert f.checked
    assert f.pricing.amount == Decimal('150.50')
    assert f.pricing.currency == 'USD'

    f = field.parseString("[ ] Priority Boarding (15.00 USD)")
    assert f.type == 'checkbox'
    assert f.label == 'Priority Boarding'
    assert not f.checked
    assert f.pricing.amount == Decimal('15.00')
    assert f.pricing.currency == 'USD'

    f = field.parseString("[ ] Discount (-5.00 USD)")
    assert f.type == 'checkbox'
    assert f.label == 'Discount'
    assert not f.checked
    assert f.pricing.amount == Decimal('-5.00')
    assert f.pricing.currency == 'USD'


def test_non_prices():
    field = radio()

    f = field.parseString("( ) Foobar (Some Information)")
    assert f.label == 'Foobar (Some Information)'

    f = field.parseString("( ) Foobar (1up)")
    assert f.label == 'Foobar (1up)'

    f = field.parseString("( ) Foobar (123.23 USD)")
    assert f.label == 'Foobar'


def test_decimal():
    field = decimal()

    assert field.parseString('123.45')[0] == Decimal('123.45')
    assert field.parseString('123')[0] == Decimal('123')
    assert field.parseString('0.5')[0] == Decimal('0.5')
    assert field.parseString('-10.0')[0] == Decimal('-10.0')


def test_currency():
    field = currency()

    assert field.parseString('CHF')[0] == 'CHF'
    assert field.parseString('usd')[0] == 'USD'
    assert field.parseString('Cny')[0] == 'CNY'


def test_integer_range():
    field = integer_range_field()

    assert field.parseString('0..10')[0] == range(0, 10)
    assert field.parseString('-10..100')[0] == range(-10, 100)
    assert field.parseString('0..-20')[0] == range(0, -20)
    assert field.parseString('-10..-20')[0] == range(-10, -20)


def test_decimal_range():
    field = decimal_range_field()

    assert field.parseString('0.00..10.00')[0] \
        == decimal_range('0.0', '10.0')
    assert field.parseString('-10.00..100.00')[0] \
        == decimal_range('-10.0', '100.0')
    assert field.parseString('0.00..-20.00')[0] \
        == decimal_range('0.0', '-20.0')
    assert field.parseString('-10.00..-20.00')[0] \
        == decimal_range('-10.0', '-20.0')


def test_code():
    field = code()

    assert field.parseString('<markdown>').syntax == 'markdown'
    assert field.parseString('<markdown>').type == 'code'



from onegov.form import _


class Snippets(object):

    fragments = (
        (_("General"), None),
        (_("Title"), '#'),
        (_("Text"), '___'),
        (_("Multiline"), '...'),
        (_("E-Mail"), '@@@'),
        (_("Website"), 'http://'),

        (_("Date and time"), None),
        (_("Date"), 'YYYY.MM.DD'),
        (_("Time"), 'HH:MM'),
        (_("Date and time"), 'YYYY.MM.DD HH:MM'),

        (_("Choices"), None),
        (_("Choice"), (
            '\n'
            '    (x) A\n'
            '    ( ) B\n'
            '    ( ) C'
        )),
        (_("Multiple Choice"), (
            '\n'
            '    [ ] A\n'
            '    [ ] B\n'
            '    [ ] C'
        )),

        (_("Files"), None),
        (_("Image"), '*.jpg|*.png|*.gif'),
        (_("Document"), '*.pdf'),

        (_("Numbers"), None),
        (_("Age"), '0..150'),
        (_("Percentage"), '0.00..100.00'),

        (_("Extended"), None),
        (_("IBAN"), '# iban'),
        (_("Swiss social security number"), '# ch.ssn'),
        (_("Swiss business identifier"), '# ch.uid'),
        (_("Swiss vat number"), '# ch.vat'),
        (_("Markdown"), '<markdown>')
    )

    def translated(self, request):
        for title, snippet in self.fragments:
            yield request.translate(title), snippet



import re
import tempfile

from cgi import FieldStorage
from gzip import GzipFile
from io import BytesIO
from onegov.core.utils import dictionary_to_binary
from onegov.form import Form
from onegov.form.fields import ChosenSelectField
from onegov.form.fields import ChosenSelectMultipleField
from onegov.form.fields import HtmlField
from onegov.form.fields import MultiCheckboxField
from onegov.form.fields import OrderedMultiCheckboxField
from onegov.form.fields import PhoneNumberField
from onegov.form.fields import UploadField
from onegov.form.validators import ValidPhoneNumber


def create_file(mimetype, filename, content):
    fs = FieldStorage()
    fs.file = tempfile.TemporaryFile("wb+")
    fs.type = mimetype
    fs.filename = filename
    fs.file.write(content)
    fs.file.seek(0)
    return fs


def test_upload_file():
    field = UploadField()
    field = field.bind(Form(), 'upload')

    textfile = create_file('text/plain', 'test.txt', b'foobar')
    data = field.process_fieldstorage(textfile)

    assert data['filename'] == 'test.txt'
    assert data['mimetype'] == 'text/plain'
    assert data['size']
    assert data['data']

    def decompress(data):
        with GzipFile(filename='', mode='r', fileobj=BytesIO(data)) as f:
            return f.read()

    assert dictionary_to_binary(data) == b'foobar'


def test_ordered_multi_checkbox_field():
    ordinary = MultiCheckboxField(choices=[
        ('c', 'C'),
        ('b', 'B'),
        ('a', 'A')
    ])
    ordered = OrderedMultiCheckboxField(choices=[
        ('c', 'C'),
        ('b', 'B'),
        ('a', 'A')
    ])
    ordinary = ordinary.bind(Form(), 'choices')
    ordered = ordered.bind(Form(), 'choices')

    ordinary.data = ordered.data = []

    assert re.findall(r'value="((a|b|c){1})"', ordinary()) == [
        ('c', 'c'),
        ('b', 'b'),
        ('a', 'a'),
    ]

    assert re.findall(r'value="((a|b|c){1})"', ordered()) == [
        ('a', 'a'),
        ('b', 'b'),
        ('c', 'c'),
    ]


def test_html_field():
    form = Form()
    field = HtmlField()
    field = field.bind(form, 'html')

    field.data = ''
    assert 'class="editor"' in field()
    assert field.validate(form)

    field.data = '<script>alert(0)</script>'
    field.validate(form)
    assert '<script>' not in field.data


def test_phone_number_field():
    form = Form()
    field = PhoneNumberField()
    field = field.bind(form, 'phone_number')
    assert field.country == 'CH'
    assert field.validators[0].country == 'CH'

    field.data = ''
    assert field.formatted_data == ''
    assert field.validate(form)

    field.data = '791112233'
    assert field.formatted_data == '+41791112233'
    assert field.validate(form)

    field.data = 'abc'
    assert field.formatted_data == 'abc'
    assert not field.validate(form)

    form = Form()
    field = PhoneNumberField(country='DE')
    field = field.bind(form, 'phone_number')
    assert field.validators[0].country == 'DE'
    assert field.country == 'DE'

    form = Form()
    field = PhoneNumberField(validators=[ValidPhoneNumber(country='DE')])
    field = field.bind(form, 'phone_number')
    assert field.validators[0].country == 'DE'


def test_chosen_select_field():
    form = Form()
    field = ChosenSelectField(choices=(('a', 'b'),))
    field = field.bind(form, 'choice')

    field.data = ''
    assert 'class="chosen-select"' in field()
    assert 'data-no_results_text="No results match"' in field()
    assert 'data-placeholder="Select an Option"' in field()


def test_chosen_select_multiple_field():
    form = Form()
    field = ChosenSelectMultipleField(choices=(('a', 'b'),))
    field = field.bind(form, 'choice')

    field.data = ''
    assert 'class="chosen-select"' in field()
    assert 'data-no_results_text="No results match"' in field()
    assert 'data-placeholder="Select Some Options"' in field()



from onegov.form.models.definition import FormDefinition
from onegov.form.models.submission import FormSubmission
from onegov.form.models.submission import PendingFormSubmission
from onegov.form.models.submission import CompleteFormSubmission
from onegov.form.models.submission import FormFile
from onegov.form.models.registration_window import FormRegistrationWindow

__all__ = (
    'FormDefinition',
    'FormRegistrationWindow',
    'FormSubmission',
    'PendingFormSubmission',
    'CompleteFormSubmission',
    'FormFile',
)



from onegov.core.orm import Base
from onegov.core.orm.mixins import ContentMixin, TimestampMixin
from onegov.core.orm.mixins import meta_property, content_property
from onegov.core.utils import normalize_for_url
from onegov.form.models.submission import FormSubmission
from onegov.form.models.registration_window import FormRegistrationWindow
from onegov.form.parser import parse_form
from onegov.form.utils import hash_definition
from onegov.form.extensions import Extendable
from onegov.search import ORMSearchable
from sqlalchemy import Column, Text
from sqlalchemy.orm import object_session, relationship
from sqlalchemy_utils import observes


class SearchableDefinition(ORMSearchable):
    """ Defines how the definitions are searchable. For now, submissions are
    not searched as they are usually accessed through the ticket, at least in
    onegov.town. If other modules need this, it can be added here and
    onegov.town can decied not to search for submissions.

    """
    es_id = 'name'
    es_public = True

    es_properties = {
        'title': {'type': 'localized'},
        'lead': {'type': 'localized'},
        'text': {'type': 'localized_html'}
    }


class FormDefinition(Base, ContentMixin, TimestampMixin, SearchableDefinition,
                     Extendable):
    """ Defines a form stored in the database. """

    __tablename__ = 'forms'

    #: the name of the form (key, part of the url)
    name = Column(Text, nullable=False, primary_key=True)

    #: the title of the form
    title = Column(Text, nullable=False)

    #: the form as parsable string
    definition = Column(Text, nullable=False)

    #: the group to which this resource belongs to (may be any kind of string)
    group = Column(Text, nullable=True)

    #: The normalized title for sorting
    order = Column(Text, nullable=False, index=True)

    #: the checksum of the definition, forms and submissions with matching
    #: checksums are guaranteed to have the exact same definition
    checksum = Column(Text, nullable=False)

    #: the type of the form, this can be used to create custom polymorphic
    #: subclasses. See `<http://docs.sqlalchemy.org/en/improve_toc/
    #: orm/extensions/declarative/inheritance.html>`_.
    type = Column(Text, nullable=True)

    #: link between forms and submissions
    submissions = relationship('FormSubmission', backref='form')

    #: link between forms and registration windows
    registration_windows = relationship(
        'FormRegistrationWindow',
        backref='form',
        order_by='FormRegistrationWindow.start')

    #: the currently active registration window
    #:
    #: this sorts the registration windows by the smaller k-nearest neighbour
    #: result of both start and end in relation to the current date
    #:
    #: the result is the *nearest* date range in relation to today:
    #: * during an active registration window, it's that active window
    #: * outside of active windows, it's last window half way until
    #:   the next window starts
    #:
    #: this could of course be done more conventionally, but this is cooler 😅
    current_registration_window = relationship(
        'FormRegistrationWindow', viewonly=True, uselist=False,
        primaryjoin="""and_(
            FormRegistrationWindow.name == FormDefinition.name,
            FormRegistrationWindow.id == select((
                FormRegistrationWindow.id,
            )).where(
                FormRegistrationWindow.name == FormDefinition.name
            ).order_by(
                func.least(
                    cast(
                        func.now().op('AT TIME ZONE')(
                            FormRegistrationWindow.timezone
                        ), Date
                    ).op('<->')(FormRegistrationWindow.start),
                    cast(
                        func.now().op('AT TIME ZONE')(
                            FormRegistrationWindow.timezone
                        ), Date
                    ).op('<->')(FormRegistrationWindow.end)
                )
            ).limit(1)
        )"""
    )

    #: lead text describing the form
    lead = meta_property()

    #: content associated with the form
    text = content_property()

    #: payment options ('manual' for out of band payments without cc, 'free'
    #: for both manual and cc payments, 'cc' for forced cc payments)
    payment_method = Column(Text, nullable=False, default='manual')

    __mapper_args__ = {
        "polymorphic_on": 'type'
    }

    @property
    def form_class(self):
        """ Parses the form definition and returns a form class. """

        return self.extend_form_class(
            parse_form(self.definition), self.meta.get('extensions'))

    @observes('definition')
    def definition_observer(self, definition):
        self.checksum = hash_definition(definition)

    @observes('title')
    def title_observer(self, title):
        self.order = normalize_for_url(title)

    def has_submissions(self, with_state=None):
        query = object_session(self).query(FormSubmission.id)
        query = query.filter(FormSubmission.name == self.name)

        if with_state is not None:
            query = query.filter(FormSubmission.state == with_state)

        return query.first() and True or False

    def add_registration_window(self, start, end, **options):
        window = FormRegistrationWindow(
            start=start,
            end=end,
            **options
        )

        self.registration_windows.append(window)

        return window



from decimal import Decimal
from onegov.form import Form, merge_forms, with_options, move_fields
from onegov.pay import Price
from wtforms import RadioField, StringField, TextAreaField, validators
from wtforms.fields.html5 import EmailField
from wtforms.validators import InputRequired
from wtforms.widgets import TextArea
from wtforms_components import TimeField


class DummyPostData(dict):
    def getlist(self, key):
        v = self[key]
        if not isinstance(v, (list, tuple)):
            v = [v]
        return v


class DummyRequest(object):
    def __init__(self, POST):
        self.POST = DummyPostData(POST)


class DummyField(object):
    def __init__(self, id, name, value):
        self.id = id
        self.name = name
        self.value = value

    def _value(self):
        return self.value


def test_submitted():

    class TestForm(Form):
        test = StringField("Test", [validators.DataRequired()])

    request = DummyRequest({})
    assert not TestForm(request.POST).submitted(request)

    request = DummyRequest({'test': 'Test'})
    assert TestForm(request.POST).submitted(request)


def test_useful_data():

    class TestForm(Form):
        a = StringField("a")
        b = StringField("b")
        c = StringField("c")

    request = DummyRequest({'a': 'A', 'b': 'B', 'c': 'C'})
    assert TestForm(request.POST).get_useful_data(exclude={'a', 'b'}) \
        == {'c': 'C'}


def test_with_options():
    widget = with_options(TextArea, class_="markdown")
    assert 'class="markdown"' in widget(DummyField('one', 'one', '1'))

    widget = with_options(TextArea, class_="x")
    assert 'class="x"' in widget(DummyField('one', 'one', '1'))


def test_match_fields():

    class TestForm(Form):
        name = StringField("Name", [validators.DataRequired()])
        email = EmailField("E-Mail")
        comment = TextAreaField("Comment")

    form = TestForm()
    assert form.match_fields(required=True) == ['name']
    assert form.match_fields(required=False) == ['email', 'comment']
    assert form.match_fields(required=None) == ['name', 'email', 'comment']
    assert form.match_fields(required=None, limit=1) == ['name']
    assert form.match_fields(include_classes=(StringField, ))\
        == ['name', 'email', 'comment']
    assert form.match_fields(include_classes=(EmailField, )) == ['email']
    assert form.match_fields(exclude_classes=(TextAreaField, ))\
        == ['name', 'email']


def test_dependent_field():

    class TestForm(Form):
        switch = RadioField(
            label="Switch",
            choices=[
                ('on', "On"),
                ('off', "Off")
            ]
        )

        optional = TimeField(
            label="Optional",
            validators=[InputRequired()],
            depends_on=('switch', 'on')
        )

    request = DummyRequest({'switch': 'off'})
    form = TestForm(request.POST)
    assert form.validate()

    request = DummyRequest({'switch': 'off', 'optional': ''})
    form = TestForm(request.POST)
    assert form.validate()

    # even though the second field is optional, it still needs to be valid
    # if there's a value (it may be empty) - we can never accept invalid values
    # as this presents a possible security risk (we could accessing something
    # that a validator filters out for security reasons).
    request = DummyRequest({'switch': 'off', 'optional': 'asdf'})
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest({'switch': 'off', 'optional': '12:00'})
    form = TestForm(request.POST)
    assert form.validate()

    request = DummyRequest({'switch': 'on'})
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest({'switch': 'on', 'optional': ''})
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest({'switch': 'on', 'optional': 'asdf'})
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest({'switch': 'on', 'optional': '12:00'})
    form = TestForm(request.POST)
    assert form.validate()


def test_dependent_field_inverted():

    class TestForm(Form):
        switch = RadioField(
            label="Switch",
            choices=[
                ('on', "On"),
                ('off', "Off")
            ]
        )

        optional = TimeField(
            label="Optional",
            validators=[InputRequired()],
            depends_on=('switch', '!off')
        )

    request = DummyRequest({'switch': 'off'})
    form = TestForm(request.POST)
    assert form.validate()

    request = DummyRequest({'switch': 'off', 'optional': ''})
    form = TestForm(request.POST)
    assert form.validate()

    # even though the second field is optional, it still needs to be valid
    # if there's a value (it may be empty) - we can never accept invalid values
    # as this presents a possible security risk (we could accessing something
    # that a validator filters out for security reasons).
    request = DummyRequest({'switch': 'off', 'optional': 'asdf'})
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest({'switch': 'off', 'optional': '12:00'})
    form = TestForm(request.POST)
    assert form.validate()

    request = DummyRequest({'switch': 'on'})
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest({'switch': 'on', 'optional': ''})
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest({'switch': 'on', 'optional': 'asdf'})
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest({'switch': 'on', 'optional': '12:00'})
    form = TestForm(request.POST)
    assert form.validate()


def test_dependent_field_multiple():

    class TestForm(Form):
        switch_1 = RadioField(
            label="Switch",
            choices=[
                ('on', "On"),
                ('off', "Off")
            ]
        )

        switch_2 = RadioField(
            label="Switch",
            choices=[
                ('on', "On"),
                ('off', "Off")
            ]
        )

        optional = TimeField(
            label="Optional",
            validators=[InputRequired()],
            depends_on=('switch_1', 'on', 'switch_2', 'off')
        )

    # optional not present
    request = DummyRequest({'switch_1': 'on', 'switch_2': 'on'})
    form = TestForm(request.POST)
    assert form.validate()
    request = DummyRequest({'switch_1': 'off', 'switch_2': 'off'})
    form = TestForm(request.POST)
    assert form.validate()
    request = DummyRequest({'switch_1': 'off', 'switch_2': 'on'})
    form = TestForm(request.POST)
    assert form.validate()

    # optional empty
    request = DummyRequest(
        {'switch_1': 'on', 'switch_2': 'on', 'optional': ''}
    )
    form = TestForm(request.POST)
    assert form.validate()
    request = DummyRequest(
        {'switch_1': 'off', 'switch_2': 'off', 'optional': ''}
    )
    form = TestForm(request.POST)
    assert form.validate()
    request = DummyRequest(
        {'switch_1': 'off', 'switch_2': 'on', 'optional': ''}
    )
    form = TestForm(request.POST)
    assert form.validate()

    # even though the second field is optional, it still needs to be valid
    # if there's a value (it may be empty) - we can never accept invalid values
    # as this presents a possible security risk (we could accessing something
    # that a validator filters out for security reasons).
    request = DummyRequest(
        {'switch_1': 'on', 'switch_2': 'on', 'optional': 'asdf'}
    )
    form = TestForm(request.POST)
    assert not form.validate()
    request = DummyRequest(
        {'switch_1': 'off', 'switch_2': 'off', 'optional': 'asdf'}
    )
    form = TestForm(request.POST)
    assert not form.validate()
    request = DummyRequest(
        {'switch_1': 'off', 'switch_2': 'on', 'optional': 'asdf'}
    )
    form = TestForm(request.POST)
    assert not form.validate()
    request = DummyRequest(
        {'switch_1': 'on', 'switch_2': 'off', 'optional': 'asdf'}
    )
    form = TestForm(request.POST)
    assert not form.validate()

    # dependency fullfilled
    request = DummyRequest(
        {'switch_1': 'on', 'switch_2': 'off'}
    )
    form = TestForm(request.POST)
    assert not form.validate()
    request = DummyRequest(
        {'switch_1': 'on', 'switch_2': 'off', 'optional': ''}
    )
    form = TestForm(request.POST)
    assert not form.validate()

    request = DummyRequest(
        {'switch_1': 'on', 'switch_2': 'off', 'optional': '12:00'}
    )
    form = TestForm(request.POST)
    assert form.validate()


def test_merge_forms():

    class Name(Form):
        form = 'Name'
        name = StringField("Name")

        def is_valid_name(self):
            return True

    class Location(Form):
        form = 'Location'
        lat = StringField("Lat")
        lon = StringField("Lat")

        def is_valid_coordinate(self):
            return True

    class User(Form):
        form = 'User'
        name = StringField("User")

        def is_valid_user(self):
            return True

    full = merge_forms(Name, Location, User)()
    assert list(full._fields.keys()) == ['name', 'lat', 'lon']
    assert full.name.label.text == "Name"
    assert full.form == 'Name'
    assert full.is_valid_name()
    assert full.is_valid_coordinate()
    assert full.is_valid_user()

    full = merge_forms(User, Location, Name)()
    assert list(full._fields.keys()) == ['name', 'lat', 'lon']
    assert full.name.label.text == "User"
    assert full.form == 'User'
    assert full.is_valid_name()
    assert full.is_valid_coordinate()
    assert full.is_valid_user()

    full = merge_forms(Location, User, Name)()
    assert list(full._fields.keys()) == ['lat', 'lon', 'name']
    assert full.name.label.text == "User"
    assert full.form == 'Location'
    assert full.is_valid_name()
    assert full.is_valid_coordinate()
    assert full.is_valid_user()


def test_move_fields():

    class Test(Form):
        a = StringField('a')
        b = StringField('b')
        c = StringField('c')

    Moved = move_fields(Test, ['a'], None)

    assert list(Test()._fields.keys()) == ['a', 'b', 'c']
    assert list(Moved()._fields.keys()) == ['b', 'c', 'a']


def test_delete_fields():

    class Test(Form):
        a = StringField('a', fieldset='Letters')
        one = StringField('1', fieldset='Numerals')
        two = StringField('2', fieldset='Numerals')

    form = Test()
    assert form.a.name == 'a'
    assert form.one.name == 'one'
    assert form.two.name == 'two'
    assert len(form.fieldsets) == 2

    form.delete_field('one')

    assert form.a.name == 'a'
    assert form.one is None
    assert form.two.name == 'two'
    assert len(form.fieldsets) == 2

    form.delete_field('two')

    assert form.a.name == 'a'
    assert form.one is None
    assert form.two is None
    assert len(form.fieldsets) == 1


def test_ensurances():

    class EnsuredForm(Form):
        foo = StringField('foo')
        bar = StringField('bar')

        ensure_foo_and_bar_called = 0
        ensure_foo_or_bar_called = 0

        @property
        def ensure_not_triggered(self):
            assert False

        def ensure_foo_and_bar(self):
            self.ensure_foo_and_bar_called += 1

            if not (self.foo.data and self.bar.data):
                return False

        def ensure_foo_or_bar(self):
            self.ensure_foo_or_bar_called += 1

            if not (self.foo.data or self.bar.data):
                return False

    form = EnsuredForm()

    assert not form.validate()
    assert form.ensure_foo_and_bar_called == 1
    assert form.ensure_foo_or_bar_called == 1

    form.foo.data = 'foo'
    form.bar.data = 'bar'

    assert form.validate()

    assert form.ensure_foo_and_bar_called == 2
    assert form.ensure_foo_or_bar_called == 2


def test_pricing():

    class TestForm(Form):
        ticket_insurance = RadioField('Option', choices=[
            ('yes', 'Yes'),
            ('no', 'No')
        ], pricing={
            'yes': (10.0, 'CHF'),
        })

        discount_code = StringField('Discount Code', pricing={
            'FOO': (-5.0, 'CHF')
        })

    def post(data):
        return DummyRequest(data).POST

    form = TestForm(post({}))
    assert form.total() is None
    assert form.prices() == []

    form = TestForm(post({'ticket_insurance': 'no'}))
    assert form.total() is None
    assert form.prices() == []

    form = TestForm(post({'ticket_insurance': 'yes'}))
    assert form.total() == Price(Decimal('10.0'), 'CHF')
    assert form.prices() == [
        ('ticket_insurance', Price(Decimal(10.0), 'CHF'))
    ]

    form = TestForm(post({'ticket_insurance': 'yes', 'discount_code': 'test'}))
    assert form.total() == Price(Decimal('10.0'), 'CHF')
    assert form.prices() == [
        ('ticket_insurance', Price(Decimal(10.0), 'CHF'))
    ]

    form = TestForm(post({'ticket_insurance': 'yes', 'discount_code': 'FOO'}))
    assert form.total() == Price(Decimal('5.0'), 'CHF')
    assert form.prices() == [
        ('ticket_insurance', Price(Decimal(10.0), 'CHF')),
        ('discount_code', Price(Decimal(-5.0), 'CHF'))
    ]


def test_dependent_pricing():

    class TestForm(Form):

        give_donation = RadioField('Option', choices=[
            ('yes', 'Yes'),
            ('no', 'No')
        ])

        donation = RadioField('Option', choices=[
            ('small', '10 CHF'),
            ('big', '100 CHF'),
        ], pricing={
            'small': (10.0, 'CHF'),
            'big': (100.0, 'CHF')
        }, depends_on=(
            'give_donation', 'yes'
        ))

    def post(data):
        return DummyRequest(data).POST

    form = TestForm(post({'give_donation': 'yes', 'donation': 'small'}))
    assert form.total() == Price(Decimal('10.0'), 'CHF')
    assert form.prices() == [
        ('donation', Price(Decimal(10.0), 'CHF'))
    ]

    form = TestForm(post({'give_donation': 'yes', 'donation': 'big'}))
    assert form.total() == Price(Decimal('100.0'), 'CHF')
    assert form.prices() == [
        ('donation', Price(Decimal(100.0), 'CHF'))
    ]

    form = TestForm(post({'give_donation': 'no', 'donation': 'small'}))
    assert form.total() is None
    assert form.prices() == []

    form = TestForm(post({'give_donation': 'no', 'donation': 'big'}))
    assert form.total() is None
    assert form.prices() == []


def test_nested_dependent_pricing():

    class TestForm(Form):

        give_donation = RadioField('Option', choices=[
            ('yes', 'Yes'),
            ('no', 'No')
        ])

        really_give = RadioField('Option', choices=[
            ('yes', 'Yes'),
            ('no', 'No')
        ], depends_on=('give_donation', 'yes'))

        donation = RadioField('Option', choices=[
            ('small', '10 CHF'),
            ('big', '100 CHF'),
        ], pricing={
            'small': (10.0, 'CHF'),
            'big': (100.0, 'CHF')
        }, depends_on=(
            'really_give', 'yes'
        ))

    def post(data):
        return DummyRequest(data).POST

    assert TestForm(post({
        'give_donation': 'no',
        'really_give': 'yes',
        'donation': 'small'
    })).total() is None

    assert TestForm(post({
        'give_donation': 'yes',
        'really_give': 'no',
        'donation': 'small'
    })).total() is None

    assert TestForm(post({
        'give_donation': 'yes',
        'really_give': 'yes',
        'donation': 'small'
    })).total() == Price(Decimal(10.0), 'CHF')


def test_clone_form():

    class FooForm(Form):
        base = StringField('123')
        name = StringField('Foo')

    class BarForm(FooForm):
        name = StringField('Bar')

    assert BarForm().base.label.text == '123'
    assert FooForm().base.label.text == '123'

    assert BarForm().name.label.text == 'Bar'
    assert FooForm().name.label.text == 'Foo'

    FooForm.name.args = ('Noo', )
    FooForm.base.args = ('Abc', )

    assert BarForm().base.label.text == 'Abc'
    assert FooForm().base.label.text == 'Abc'

    assert BarForm().name.label.text == 'Bar'
    assert FooForm().name.label.text == 'Noo'

    NewForm = FooForm.clone()
    NewForm.base.args = ('Xyz', )

    assert BarForm().base.label.text == 'Abc'
    assert FooForm().base.label.text == 'Abc'
    assert NewForm().base.label.text == 'Xyz'

    assert BarForm().name.label.text == 'Bar'
    assert FooForm().name.label.text == 'Noo'
    assert NewForm().name.label.text == 'Noo'



import html

from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON, UUID
from onegov.core.orm.types import UTCDateTime
from onegov.file import AssociatedFiles, File
from onegov.form.display import render_field
from onegov.form.extensions import Extendable
from onegov.form.parser import parse_form
from onegov.form.utils import extract_text_from_html, hash_definition
from onegov.pay import Payable
from onegov.pay import process_payment
from sedate import utcnow
from sqlalchemy import case
from sqlalchemy import CheckConstraint
from sqlalchemy import Column
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy_utils import observes
from uuid import uuid4
from wtforms import StringField, TextAreaField
from wtforms.fields.html5 import EmailField


class FormSubmission(Base, TimestampMixin, Payable, AssociatedFiles,
                     Extendable):
    """ Defines a submitted form in the database. """

    __tablename__ = 'submissions'

    #: id of the form submission
    id = Column(UUID, primary_key=True, default=uuid4)

    #: name of the form this submission belongs to
    name = Column(Text, ForeignKey("forms.name"), nullable=True)

    #: the title of the submission, generated from the submitted fields
    #: NULL for submissions which are not complete
    title = Column(Text, nullable=True)

    #: the e-mail address associated with the submitee, generated from the
    # submitted fields (may be NULL, even for complete submissions)
    email = Column(Text, nullable=True)

    #: the source code of the form at the moment of submission. This is stored
    #: alongside the submission as the original form may change later. We
    #: want to keep the old form around just in case.
    definition = Column(Text, nullable=False)

    #: the exact time this submissions was changed from 'pending' to 'complete'
    received = Column(UTCDateTime, nullable=True)

    #: the checksum of the definition, forms and submissions with matching
    #: checksums are guaranteed to have the exact same definition
    checksum = Column(Text, nullable=False)

    #: metadata about this submission
    meta = Column(JSON, nullable=False)

    #: the submission data
    data = Column(JSON, nullable=False)

    #: the state of the submission
    state = Column(
        Enum('pending', 'complete', name='submission_state'),
        nullable=False
    )

    #: the number of spots this submission wants to claim
    #: (only relevant if there's a registration window)
    spots = Column(Integer, nullable=False, default=0)

    #: the number of spots this submission has actually received
    #: None => the decision if spots should be given is still open
    #: 0 => the decision was negative, no spots were given
    #: 1-x => the decision was positive, at least some spots were given
    claimed = Column(Integer, nullable=True, default=None)

    #: the registration window linked with this submission
    registration_window_id = Column(
        UUID, ForeignKey("registration_windows.id"), nullable=True)

    #: payment options -> copied from the dfinition at the moment of
    #: submission. This is stored alongside the submission as the original
    #: form setting may change later.
    payment_method = Column(Text, nullable=False, default='manual')

    __mapper_args__ = {
        "polymorphic_on": 'state'
    }

    __table_args__ = (
        CheckConstraint(
            'COALESCE(claimed, 0) <= spots',
            name='claimed_no_more_than_requested'
        ),
    )

    @property
    def payable_reference(self):
        if self.name:
            return f'{self.name}/{self.title}@{self.received.isoformat()}'
        else:
            return f'{self.title}@{self.received.isoformat()}'

    @property
    def form_class(self):
        """ Parses the form definition and returns a form class. """

        return self.extend_form_class(
            parse_form(self.definition), self.meta.get('extensions'))

    @property
    def form_obj(self):
        """ Returns a form instance containing the submission data. """

        return self.form_class(data=self.data)

    @hybrid_property
    def registration_state(self):
        if not self.spots:
            return None
        if self.claimed is None:
            return 'open'
        elif self.claimed == 0:
            return 'cancelled'
        elif self.claimed == self.spots:
            return 'confirmed'
        elif self.claimed < self.spots:
            return 'partial'

    @registration_state.expression
    def registration_state(cls):
        return case((
            (cls.spots == 0, None),
            (cls.claimed == None, 'open'),
            (cls.claimed == 0, 'cancelled'),
            (cls.claimed == cls.spots, 'confirmed'),
            (cls.claimed < cls.spots, 'partial')
        ), else_=None)

    def get_email_field_data(self, form=None):
        form = form or self.form_obj

        email_fields = form.match_fields(
            include_classes=(EmailField, ),
            required=True,
            limit=1
        )
        email_fields += form.match_fields(
            include_classes=(EmailField, ),
            required=False,
            limit=1
        )

        if email_fields:
            return form._fields[email_fields[0]].data

    @observes('definition')
    def definition_observer(self, definition):
        self.checksum = hash_definition(definition)

    @observes('state')
    def state_observer(self, state):
        if self.state == 'complete':
            form = self.form_class(data=self.data)

            title_fields = form.match_fields(
                include_classes=(StringField, ),
                exclude_classes=(TextAreaField, ),
                required=True,
                limit=3
            )

            if title_fields:
                self.title = extract_text_from_html(', '.join(
                    html.unescape(render_field(form._fields[id]))
                    for id in title_fields
                ))

            if not self.email:
                self.email = self.get_email_field_data(form=form)

            # only set the date the first time around
            if not self.received:
                self.received = utcnow()

    def process_payment(self, price, provider=None, token=None):
        """ Takes a request, optionally with the provider and the token
        by the provider that can be used to charge the credit card and creates
        a payment record if necessary.

        Returns True or a payment object if the payment was processed
        successfully. That is, if there is a payment or if there is no payment
        required the method returns truthy.

        """

        if price and price.amount > 0:
            return process_payment(self.payment_method, price, provider, token)

        return True

    def claim(self, spots=None):
        """ Claimes the given number of spots (defaults to the requested
        number of spots).

        """
        spots = spots or self.spots

        assert self.registration_window

        if self.registration_window.limit:
            limit = self.registration_window.limit
            claimed = self.registration_window.claimed_spots

            assert spots <= (limit - claimed)

        assert ((self.claimed or 0) + spots) <= self.spots
        self.claimed = (self.claimed or 0) + spots

    def disclaim(self, spots=None):
        """ Disclaims the given number of spots (defaults to all spots that
        were claimed so far).

        """
        spots = spots or self.spots

        assert self.registration_window

        if self.claimed is None:
            self.claimed = 0
        else:
            self.claimed = max(0, self.claimed - spots)


class PendingFormSubmission(FormSubmission):
    __mapper_args__ = {'polymorphic_identity': 'pending'}


class CompleteFormSubmission(FormSubmission):
    __mapper_args__ = {'polymorphic_identity': 'complete'}


class FormFile(File):
    __mapper_args__ = {'polymorphic_identity': 'formfile'}



from onegov.form.validators import ValidPhoneNumber
from pytest import raises
from wtforms.validators import ValidationError


def test_phone_number_validator():

    class Field(object):
        def __init__(self, data):
            self.data = data

    validator = ValidPhoneNumber()

    validator(None, Field(None))
    validator(None, Field(''))

    validator(None, Field('+41791112233'))
    validator(None, Field('0041791112233'))
    validator(None, Field('0791112233'))

    with raises(ValidationError):
        validator(None, Field(1234))
    with raises(ValidationError):
        validator(None, Field('1234'))

    with raises(ValidationError):
        validator(None, Field('+417911122333'))
    with raises(ValidationError):
        validator(None, Field('041791112233'))
    with raises(ValidationError):
        validator(None, Field('041791112233'))
    with raises(ValidationError):
        validator(None, Field('00791112233'))



import pytest
import transaction

from datetime import date, timedelta
from onegov.form import FormCollection, FormExtension, FormSubmission
from sqlalchemy import desc
from sqlalchemy.exc import IntegrityError
from webob.multidict import MultiDict
from wtforms import ValidationError


def days(d):
    return timedelta(days=d)


def test_has_submissions(session):
    collection = FormCollection(session)
    form = collection.definitions.add('Newsletter', definition="E-Mail = @@@")

    assert not form.has_submissions()

    data = MultiDict([
        ('e_mail', 'test@example.org'),
    ])

    collection.submissions.add(
        'newsletter', form.form_class(data), state='complete')

    assert not form.has_submissions(with_state='pending')
    assert form.has_submissions(with_state='complete')
    assert form.has_submissions()


def test_form_extensions(session):
    collection = FormCollection(session)

    members = collection.definitions.add('Members', definition="E-Mail = @@@")

    class CorporateOnly(object):

        def validate_e_mail(self, value):
            # note, you probably don't want to do this in a real world
            # project as the name of the e-mail field might change and
            # the validation would not be triggered!
            if not value.data.endswith('seantis.ch'):
                raise ValidationError("Only seantis.ch e-mails are allowed")

    class CorporateOnlyExtension(FormExtension, name='corporate-emails-only'):

        def create(self):
            class ExtendedForm(self.form_class, CorporateOnly):
                pass

            return ExtendedForm

    members.meta['extensions'] = ['corporate-emails-only']
    assert issubclass(members.form_class, CorporateOnly)

    session.flush()

    # the validation should fail here
    data = MultiDict([
        ('e_mail', 'test@example.org'),
    ])

    form = members.form_class(data)
    form.validate()
    assert form.errors

    with pytest.raises(AssertionError) as e:
        collection.submissions.add('newsletter', form, state='complete')

    assert "the given form doesn't validate" in str(e)

    # now it should work
    data = MultiDict([
        ('e_mail', 'test@seantis.ch'),
    ])

    form = members.form_class(data)
    form.validate()
    assert not form.errors

    # the extensions are carried over to the submission
    submission = collection.submissions.add(
        'members', form, state='complete')
    assert issubclass(submission.form_class, CorporateOnly)


def test_registration_window_adjacent(session):
    forms = FormCollection(session)

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")
    winter = forms.definitions.add('Witnercamp', definition="E-Mail = @@@")

    summer.add_registration_window(date(2017, 4, 1), date(2017, 6, 30))
    summer.add_registration_window(date(2018, 4, 1), date(2018, 6, 30))

    assert len(summer.registration_windows) == 2

    winter.add_registration_window(date(2017, 4, 1), date(2017, 6, 30))
    winter.add_registration_window(date(2018, 4, 1), date(2018, 6, 30))

    assert len(winter.registration_windows) == 2

    # no overlap -> different forms
    session.flush()

    # adjacent, fails
    summer.add_registration_window(date(2017, 1, 1), date(2017, 4, 1))

    with pytest.raises(IntegrityError) as e:
        session.flush()

    assert 'no_adjacent_registration_windows' in str(e)


def test_registration_window_overlaps(session):
    forms = FormCollection(session)

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")
    summer.add_registration_window(date(2017, 4, 1), date(2017, 6, 30))

    assert len(summer.registration_windows) == 1

    # no overlap -> different forms
    session.flush()

    # overlapping, fails
    summer.add_registration_window(date(2017, 1, 1), date(2017, 4, 2))

    with pytest.raises(IntegrityError) as e:
        session.flush()

    assert 'no_overlapping_registration_windows' in str(e)


def test_registration_window_end_before_start(session):
    camp = FormCollection(session).definitions.add(
        'Camp', definition="E-Mail = @@@")

    camp.add_registration_window(date(2018, 1, 1), date(2017, 1, 1))

    with pytest.raises(IntegrityError) as e:
        session.flush()

    assert 'start_smaller_than_end' in str(e)


def test_current_registration_window_bound_to_form(session):
    forms = FormCollection(session)
    today = date.today()

    winter = forms.definitions.add('Witnercamp', definition="E-Mail = @@@")
    winter.add_registration_window(today - days(1), today + days(1))

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")
    summer.add_registration_window(today - days(100), today - days(10))

    assert winter.current_registration_window.start == today - days(1)
    assert summer.current_registration_window.start == today - days(100)


def test_current_registration_window_end_date(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")

    # the first window is closer, though the start is further away
    summer.add_registration_window(today - days(10), today - days(1))
    summer.add_registration_window(today + days(5), today + days(10))

    assert summer.current_registration_window.start == today - days(10)


def test_registration_window_spots(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")
    window = summer.add_registration_window(today - days(5), today + days(5))

    session.flush()

    window.enabled = False
    assert not window.accepts_submissions()

    window.enabled = True
    window.end = today - days(1)
    assert not window.accepts_submissions()

    window.start = today + days(1)
    window.end = today + days(5)
    assert not window.accepts_submissions()

    window.start = today - days(5)
    window.overflow = True
    assert window.accepts_submissions()

    window.overflow = False
    window.limit = None
    assert window.accepts_submissions()
    assert window.claimed_spots == 0
    assert window.requested_spots == 0

    window.limit = 2
    assert window.accepts_submissions()
    assert window.claimed_spots == 0
    assert window.requested_spots == 0

    s1 = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=1
    )

    assert window.accepts_submissions()
    assert window.claimed_spots == 0
    assert window.requested_spots == 1

    s2 = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=1
    )

    assert not window.accepts_submissions()
    assert window.claimed_spots == 0
    assert window.requested_spots == 2

    window.overflow = True
    assert window.accepts_submissions()

    s1.claim(1)
    session.flush()

    assert window.accepts_submissions()
    assert window.claimed_spots == 1
    assert window.requested_spots == 1

    window.overflow = False
    assert not window.accepts_submissions()

    s2.claim(1)
    session.flush()

    assert not window.accepts_submissions()
    assert window.claimed_spots == 2
    assert window.requested_spots == 0

    window.overflow = True
    assert window.accepts_submissions()


def test_registration_claims_with_no_limit(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")

    window = summer.add_registration_window(today - days(5), today + days(5))
    window.limit = None

    session.flush()

    submission = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=100
    )

    submission.claim()
    session.flush()

    assert submission.claimed == 100
    assert submission.spots == 100
    assert window.claimed_spots == 100
    assert window.requested_spots == 0

    submission.disclaim()
    session.flush()

    assert submission.claimed == 0
    assert submission.spots == 100
    assert window.claimed_spots == 0
    assert window.requested_spots == 0

    submission.claim(50)
    session.flush()

    assert submission.claimed == 50
    assert submission.spots == 100
    assert window.claimed_spots == 50
    assert window.requested_spots == 50

    submission.claim(25)
    session.flush()

    assert submission.claimed == 75
    assert submission.spots == 100
    assert window.claimed_spots == 75
    assert window.requested_spots == 25

    submission.claim(25)
    session.flush()

    assert submission.claimed == 100
    assert submission.spots == 100
    assert window.claimed_spots == 100
    assert window.requested_spots == 0

    submission.disclaim(25)
    session.flush()

    assert submission.claimed == 75
    assert submission.spots == 100
    assert window.claimed_spots == 75
    assert window.requested_spots == 25

    submission.disclaim()
    session.flush()

    assert submission.claimed == 0
    assert submission.spots == 100
    assert window.claimed_spots == 0
    assert window.requested_spots == 0


def test_registration_claims_with_a_limit(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")

    window = summer.add_registration_window(today - days(5), today + days(5))
    window.limit = 10
    window.overflow = True

    session.flush()

    submission = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=100
    )

    with pytest.raises(AssertionError):
        submission.claim()

    with pytest.raises(AssertionError):
        submission.claim(11)

    submission.claim(spots=10)
    session.flush()

    assert window.claimed_spots == 10
    assert window.requested_spots == 90
    assert window.available_spots == 0

    submission.disclaim()
    session.flush()

    assert window.claimed_spots == 0
    assert window.requested_spots == 0
    assert window.available_spots == 10


def test_register_more_than_allowed(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")

    window = summer.add_registration_window(today - days(5), today + days(5))
    window.limit = 1
    window.overflow = False

    session.flush()

    with pytest.raises(AssertionError):
        forms.submissions.add(
            name='summercamp',
            form=summer.form_class(data={'e_mail': 'info@example.org'}),
            state='complete',
            spots=2
        )

    window.overflow = True
    session.flush()

    forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=2
    )

    window.overflow = False
    session.flush()


def test_undo_registration(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")

    window = summer.add_registration_window(today - days(5), today + days(5))
    window.limit = 1
    window.overflow = False

    session.flush()

    assert window.available_spots == 1

    submission = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=1
    )
    session.flush()

    assert window.available_spots == 0

    submission.claim()
    session.flush()

    assert window.available_spots == 0

    submission.disclaim()
    session.flush()

    assert window.available_spots == 1


def test_registration_window_queue(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")
    window = summer.add_registration_window(today - days(5), today + days(5))

    session.flush()

    first = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=10
    )
    session.flush()

    assert window.next_submission is first

    second = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=10
    )
    session.flush()

    assert window.next_submission is first

    first.claim()
    session.flush()

    assert window.next_submission is second

    second.claim(5)
    session.flush()

    assert window.next_submission is second

    second.claim(5)
    session.flush()

    assert window.next_submission is None


def test_require_spots_if_registration_window(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")
    summer.add_registration_window(today - days(5), today + days(5))

    session.flush()

    with pytest.raises(AssertionError):
        forms.submissions.add(
            name='summercamp',
            form=summer.form_class(data={'e_mail': 'info@example.org'}),
            state='complete',
            spots=0
        )

    with pytest.raises(TypeError):
        forms.submissions.add(
            name='summercamp',
            form=summer.form_class(data={'e_mail': 'info@example.org'}),
            state='complete',
            spots=None
        )


def test_registration_submission_state(session):
    forms = FormCollection(session)
    today = date.today()

    summer = forms.definitions.add('Summercamp', definition="E-Mail = @@@")
    session.flush()

    submission = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
    )

    def query_registration_state():
        q = forms.submissions.query()
        q = q.with_entities(FormSubmission.registration_state)
        q = q.order_by(desc(FormSubmission.created))

        return q.first().registration_state

    assert submission.registration_state is None
    assert query_registration_state() is None

    summer = forms.definitions.by_name('summercamp')
    summer.add_registration_window(today - days(5), today + days(5), limit=10)

    session.flush()

    # required here for some reason, otherwise there's no current window
    transaction.commit()

    summer = forms.definitions.by_name('summercamp')
    submission = forms.submissions.add(
        name='summercamp',
        form=summer.form_class(data={'e_mail': 'info@example.org'}),
        state='complete',
        spots=2
    )
    session.flush()

    assert submission.registration_state == 'open'
    assert query_registration_state() == 'open'

    submission.claim(1)

    assert submission.registration_state == 'partial'
    assert query_registration_state() == 'partial'

    submission.claim(1)
    assert submission.registration_state == 'confirmed'
    assert query_registration_state() == 'confirmed'

    submission.disclaim()
    assert submission.registration_state == 'cancelled'
    assert query_registration_state() == 'cancelled'



from decimal import Decimal
from onegov.form import decimal_range


def test_decimal_range():
    assert list(decimal_range('0', '2')) == [
        Decimal('0'),
        Decimal('1')
    ]

    assert list(decimal_range('2', '0')) == [
        Decimal('2'),
        Decimal('1')
    ]

    assert list(decimal_range('0', '1', '0.25')) == [
        Decimal('0'),
        Decimal('0.25'),
        Decimal('0.5'),
        Decimal('0.75')
    ]

    assert decimal_range('0', '1') == decimal_range('0', '1')
    assert decimal_range('1', '0') != decimal_range('0', '1')
    assert decimal_range('0', '1', '0.1') != decimal_range('0', '1', '0.2')

    assert repr(decimal_range('0', '1')) == "decimal_range('0', '1')"



import chameleon
import humanize

from cgi import escape
from onegov.form import _
from wtforms.widgets import FileInput
from wtforms.widgets import ListWidget
from wtforms.widgets import Select
from wtforms.widgets import TextInput
from wtforms.widgets.core import HTMLString


class OrderedListWidget(ListWidget):
    """ Extends the default list widget with automated ordering using the
    translated text of each element.

    """

    def __call__(self, field, **kwargs):

        # ListWidget expects a field internally, but it will only use
        # its id property and __iter__ method, so we can get away
        # with passing a fake field with an id and an iterator.
        #
        # It's not great, since we have to assume internal knowledge,
        # but builting a new field or changing the existing one would
        # require even more knowledge, so this is the better approach
        #
        # We also need to call each field once so it gets hooked up with
        # our translation machinary
        ordered = [subfield for subfield in field]
        ordered.sort(key=lambda f: (f(), str(f.label.text))[1])

        class FakeField(object):

            id = field.id

            def __iter__(self):
                return iter(ordered)

        return super().__call__(FakeField(), **kwargs)


class MultiCheckboxWidget(ListWidget):
    """ The default list widget with the label behind the checkbox. """

    def __init__(self, *args, **kwargs):
        kwargs['prefix_label'] = False
        super().__init__(*args, **kwargs)


class OrderedMultiCheckboxWidget(MultiCheckboxWidget, OrderedListWidget):
    """ The sorted list widget with the label behind the checkbox. """
    pass


class CoordinateWidget(TextInput):
    """ Widget containing the coordinates for the
    :class:`onegov.form.fields.CoordinateField` class.

    Basically a text input with a class. Meant to be enhanced on the browser
    using javascript.

    """

    def __call__(self, field, **kwargs):
        kwargs['class_'] = (kwargs.get('class_', '') + ' coordinate').strip()
        return super().__call__(field, **kwargs)


class UploadWidget(FileInput):
    """ An upload widget for the :class:`onegov.form.fields.UploadField` class,
    which supports keeping, removing and replacing already uploaded files.

    This is necessary as file inputs are read-only on the client and it's
    therefore rather easy for users to lose their input otherwise (e.g. a
    form with a file is rejected because of some mistake - the file disappears
    once the response is rendered on the client).

    """

    def __call__(self, field, **kwargs):
        force_simple = kwargs.pop('force_simple', False)
        input_html = super().__call__(field, **kwargs)

        if force_simple or field.errors or not field.data:
            return HTMLString("""
                <div class="upload-widget without-data">
                    {}
                </div>
            """.format(input_html))
        else:
            return HTMLString("""
                <div class="upload-widget with-data">
                    <p>{existing_file_label}: {filename} ({filesize}) ✓</p>
                    <ul>
                        <li>
                            <input type="radio" id="{name}-0" name="{name}"
                                   value="keep" checked="">
                            <label for="{name}-0">{keep_label}</label>
                        </li>
                        <li>
                            <input type="radio" id="{name}-1" name="{name}"
                                   value="delete">
                            <label for="{name}-1">{delete_label}</label>
                        </li>
                        <li>
                            <input type="radio" id="{name}-2" name="{name}"
                                   value="replace">
                            <label for="{name}-2">{replace_label}</label>
                            <div>
                                <label>
                                    <div data-depends-on="{name}/replace"
                                         data-hide-label="false">
                                        {input_html}
                                    </div>
                                </label>
                            </div>
                        </li>
                    </ul>
                </div>
            """.format(
                # be careful, we do our own html generation here without any
                # safety harness - we need to carefully escape values the user
                # might supply
                filesize=humanize.naturalsize(field.data['size']),
                filename=escape(field.data['filename'], quote=True),
                name=field.id,
                input_html=input_html,
                existing_file_label=field.gettext(_('Uploaded file')),
                keep_label=field.gettext(_('Keep file')),
                delete_label=field.gettext(_('Delete file')),
                replace_label=field.gettext(_('Replace file'))
            ))


class TagsWidget(TextInput):
    # for use with https://github.com/developit/tags-input
    input_type = 'tags'


class IconWidget(TextInput):

    iconfont = 'FontAwesome'
    icons = (
        '&#xf111',  # fa-circle
        '&#xf005',  # fa-star
        '&#xf06a',  # fa-exclamation-circle
        '&#xf059',  # fa-question-circle
        '&#xf05e',  # fa-ban
        '&#xf1b9',  # fa-car
        '&#xf238',  # fa-train
        '&#xf206',  # fa-bicycle
        '&#xf291',  # fa-shopping-basket
        '&#xf1b0',  # fa-paw
        '&#xf1ae',  # fa-child
        '&#xf06d',  # fa-fire
        '&#xf1f8',  # fa-trash
        '&#xf236',  # fa-hotel
        '&#xf0f4',  # fa-coffee
        '&#xf017',  # fa-click
    )

    template = chameleon.PageTemplate("""
        <div class="icon-widget">
            <ul style="font-family: ${iconfont}">
                <li
                    tal:repeat="icon icons"
                    tal:content="structure icon"
                />
            </ul>

            <input type="hidden" name="${id}" value="${structure: value}">
        </div>
    """)

    def __call__(self, field, **kwargs):
        iconfont = kwargs.pop('iconfont', self.iconfont)
        icons = kwargs.pop('icons', self.icons)

        return HTMLString(self.template.render(
            iconfont=iconfont,
            icons=icons,
            id=field.id,
            value=field.data or icons[0]
        ))


class ChosenSelectWidget(Select):

    def __call__(self, field, **kwargs):
        kwargs['class_'] = '{} chosen-select'.format(
            kwargs.get('class_', '')
        ).strip()
        kwargs['data-placeholder'] = field.gettext(_("Select an Option"))
        kwargs['data-no_results_text'] = field.gettext(_("No results match"))
        if self.multiple:
            kwargs['data-placeholder'] = field.gettext(
                _("Select Some Options")
            )

        return super(ChosenSelectWidget, self).__call__(field, **kwargs)


class PreviewWidget(object):
    """ A widget that displays the html of a specific view whenver there's
    a change in other fields. JavaScript is used to facilitate this.

    """

    template = chameleon.PageTemplate("""
        <div class="form-preview-widget"
             data-url="${url or ''}"
             data-fields="${','.join(fields)}"
             data-events="${','.join(events)}">
        </div>
    """)

    def __call__(self, field, **kwargs):
        field.meta.request.include('preview-widget-handler')

        return HTMLString(self.template.render(
            url=callable(field.url) and field.url(field.meta) or field.url,
            fields=field.fields,
            events=field.events,
        ))



from html import escape
from onegov.form import errors
from onegov.form.core import FieldDependency
from onegov.form.core import Form
from onegov.form.fields import MultiCheckboxField
from onegov.form.fields import UploadField
from onegov.form.parser.core import parse_formcode
from onegov.form.utils import as_internal_id
from onegov.form.utils import with_options
from onegov.form.validators import ExpectedExtensions
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import Stdnum
from onegov.form.validators import StrictOptional
from wtforms import PasswordField
from wtforms import RadioField
from wtforms import StringField
from wtforms import TextAreaField
from wtforms.fields.html5 import DateField
from wtforms.fields.html5 import DateTimeLocalField
from wtforms.fields.html5 import DecimalField
from wtforms.fields.html5 import EmailField
from wtforms.fields.html5 import IntegerField
from wtforms.fields.html5 import URLField
from wtforms.validators import DataRequired
from wtforms.validators import Length
from wtforms.validators import NumberRange
from wtforms.validators import Regexp
from wtforms.validators import URL
from wtforms.widgets import TextArea
from wtforms_components import Email, If, TimeField


# increasing the default filesize is *strongly discouarged*, as we are not
# storing those files in the database, so they need to fit in memory
#
# if this value must be higher, we need to store the files outside the
# database
#
MEGABYTE = 1000 ** 2
DEFAULT_UPLOAD_LIMIT = 5 * MEGABYTE


def parse_form(text, base_class=Form):
    """ Takes the given form text, parses it and returns a WTForms form
    class (not an instance of it).

    """

    builder = WTFormsClassBuilder(base_class)

    for fieldset in parse_formcode(text):
        builder.set_current_fieldset(fieldset.label)

        for field in fieldset.fields:
            handle_field(builder, field)

    form_class = builder.form_class
    form_class._source = text

    return form_class


def handle_field(builder, field, dependency=None):
    """ Takes the given parsed field and adds it to the form. """

    if field.type == 'text':

        if field.maxlength:
            validators = [Length(max=field.maxlength)]
        else:
            validators = []

        if field.regex:
            validators.append(Regexp(field.regex))

        builder.add_field(
            field_class=StringField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            validators=validators
        )

    elif field.type == 'textarea':
        builder.add_field(
            field_class=TextAreaField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            widget=with_options(TextArea, rows=field.rows)
        )

    elif field.type == 'password':
        builder.add_field(
            field_class=PasswordField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required
        )

    elif field.type == 'email':
        builder.add_field(
            field_class=EmailField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            validators=[Email()]
        )

    elif field.type == 'url':
        builder.add_field(
            field_class=URLField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            validators=[URL()]
        )

    elif field.type == 'stdnum':
        builder.add_field(
            field_class=StringField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            validators=[Stdnum(field.format)]
        )

    elif field.type == 'date':
        builder.add_field(
            field_class=DateField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
        )

    elif field.type == 'datetime':
        builder.add_field(
            field_class=DateTimeLocalField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
        )

    elif field.type == 'time':
        builder.add_field(
            field_class=TimeField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required
        )

    elif field.type == 'fileinput':
        builder.add_field(
            field_class=UploadField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            validators=[
                ExpectedExtensions(field.extensions),
                FileSizeLimit(DEFAULT_UPLOAD_LIMIT)
            ]
        )

    elif field.type == 'radio':
        builder.add_field(
            field_class=RadioField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            choices=[(c.key, c.label) for c in field.choices],
            default=next((c.key for c in field.choices if c.selected), None),
            pricing=field.pricing,
            # do not coerce None into 'None'
            coerce=lambda v: str(v) if v is not None else v
        )

    elif field.type == 'checkbox':
        builder.add_field(
            field_class=MultiCheckboxField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            choices=[(c.key, c.label) for c in field.choices],
            default=[c.key for c in field.choices if c.selected],
            pricing=field.pricing,
            # do not coerce None into 'None'
            coerce=lambda v: str(v) if v is not None else v
        )

    elif field.type == 'integer_range':
        builder.add_field(
            field_class=IntegerField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            validators=[
                NumberRange(
                    field.range.start,
                    field.range.stop
                )
            ]
        )

    elif field.type == 'decimal_range':
        builder.add_field(
            field_class=DecimalField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            validators=[
                NumberRange(
                    field.range.start,
                    field.range.stop
                )
            ]
        )

    elif field.type == 'code':
        builder.add_field(
            field_class=TextAreaField,
            field_id=field.id,
            label=field.label,
            dependency=dependency,
            required=field.required,
            render_kw={'data-editor': field.syntax}
        )

    else:
        raise NotImplementedError

    if field.type in ('radio', 'checkbox'):
        for choice in field.choices:
            if not choice.fields:
                continue

            dependency = FieldDependency(field.id, choice.label)

            for choice_field in choice.fields:
                handle_field(builder, choice_field, dependency)


class WTFormsClassBuilder(object):
    """ Helps dynamically build a wtforms class from parsed blocks.

    For example::

        builder = WTFormsClassBuilder(BaseClass)
        builder.add_field(StringField, label='Name', required=True)

        MyForm = builder.form_class
    """

    def __init__(self, base_class):

        class DynamicForm(base_class):
            pass

        self.form_class = DynamicForm
        self.current_fieldset = None

    def set_current_fieldset(self, label):
        self.current_fieldset = label

    def validators_extend(self, validators, required, dependency):
        if required:
            if dependency is None:
                self.validators_add_required(validators)
            else:
                self.validators_add_dependency(validators, dependency)
        else:
            self.validators_add_optional(validators)

    def validators_add_required(self, validators):
        # we use the DataRequired check instead of InputRequired, since
        # InputRequired only works if the data comes over the wire. We
        # also want to load forms with data from the database, where
        # InputRequired will fail, but DataRequired will not.
        #
        # As a consequence, falsey values can't be submitted for now.
        validators.insert(0, DataRequired())

    def validators_add_dependency(self, validators, dependency):

        # if the dependency is not fulfilled, the field may be empty
        # but it must still validate otherwise (invalid = nok, empty = ok)
        validator = If(dependency.unfulfilled, StrictOptional())
        validator.field_flags = ('required', )
        validators.insert(0, validator)

        # if the dependency is fulfilled, the field is required
        validator = If(dependency.fulfilled, DataRequired())
        validator.field_flags = ('required', )
        validators.insert(0, validator)

    def validators_add_optional(self, validators):
        validators.insert(0, StrictOptional())

    def mark_as_dependent(self, field_id, dependency):
        field = getattr(self.form_class, field_id)
        widget = field.kwargs.get('widget', field.field_class.widget)

        field.kwargs['widget'] = with_options(
            widget, **dependency.html_data
        )

    def get_unique_field_id(self, label, dependency):
        # try to find a smart field_id that contains the dependency or the
        # current fieldset name - if all fails, an error will be thrown,
        # as field_ids *need* to be unique
        if dependency:
            field_id = dependency.field_id + '_' + as_internal_id(label)
        elif self.current_fieldset:
            field_id = as_internal_id(self.current_fieldset + ' ' + label)
        else:
            field_id = as_internal_id(label)

        if hasattr(self.form_class, field_id):
            raise errors.DuplicateLabelError(label=label)

        return field_id

    def add_field(self, field_class, field_id, label, required,
                  dependency=None, pricing=None, **kwargs):
        validators = kwargs.pop('validators', [])

        if hasattr(self.form_class, field_id):
            raise errors.DuplicateLabelError(label=label)

        # labels in wtforms are not escaped correctly - for safety we make sure
        # that the label is properly html escaped. See also:
        # https://github.com/wtforms/wtforms/issues/315
        # -> quotes are allowed because the label is rendered between tags,
        # not as part of the attributes
        label = escape(label, quote=False)

        self.validators_extend(validators, required, dependency)

        setattr(self.form_class, field_id, field_class(
            label=label,
            validators=validators,
            fieldset=self.current_fieldset,
            pricing=pricing,
            **kwargs
        ))

        if dependency:
            self.mark_as_dependent(field_id, dependency)

        return getattr(self.form_class, field_id)



import importlib
import humanize
import phonenumbers

from cgi import FieldStorage
from mimetypes import types_map
from onegov.form import _
from onegov.form.utils import with_options
from onegov.form.errors import InvalidFormSyntax, DuplicateLabelError
from stdnum.exceptions import ValidationError as StdnumValidationError
from wtforms import ValidationError
from wtforms.fields import SelectField
from wtforms.validators import StopValidation, Optional
from wtforms.compat import string_types


class Stdnum(object):
    """ Validates a string using any python-stdnum format.

    See `<https://github.com/arthurdejong/python-stdnum>`_.

    """
    def __init__(self, format):
        module = '.'.join(p for p in format.split('.') if p)
        self.format = importlib.import_module('stdnum.' + module)

    def __call__(self, form, field):
        # only do a check for filled out values, to check for the existance
        # of any value use DataRequired!
        if not field.data:
            return

        try:
            self.format.validate(field.data)
        except StdnumValidationError:
            raise ValidationError(field.gettext('Invalid input.'))


class FileSizeLimit(object):
    """ Makes sure an uploaded file is not bigger than the given number of
    bytes.

    Expects an :class:`onegov.form.fields.UploadField` instance.

    """

    message = _(
        "The file is too large, please provide a file smaller than {}."
    )

    def __init__(self, max_bytes):
        self.max_bytes = max_bytes

    def __call__(self, form, field):
        if field.data:
            if field.data.get('size', 0) > self.max_bytes:
                message = field.gettext(self.message).format(
                    humanize.naturalsize(self.max_bytes)
                )
                raise ValidationError(message)


class WhitelistedMimeType(object):
    """ Makes sure an uploaded file is in a whitelist of allowed mimetypes.

    Expects an :class:`onegov.form.fields.UploadField` instance.
    """

    whitelist = {
        'application/excel',
        'application/vnd.ms-excel',
        'application/msword',
        'application/pdf',
        'application/zip',
        'image/gif',
        'image/jpeg',
        'image/png',
        'image/x-ms-bmp',
        'text/plain',
    }

    message = _("Files of this type are not supported.")

    def __init__(self, whitelist=None):
        if whitelist is not None:
            self.whitelist = whitelist

    def __call__(self, form, field):
        if field.data:
            if field.data['mimetype'] not in self.whitelist:
                raise ValidationError(field.gettext(self.message))


class ExpectedExtensions(WhitelistedMimeType):
    """ Makes sure an uploaded file has one of the expected extensions. Since
    extensions are not something we can count on we look up the mimetype of
    the extension and use that to check.

    Expects an :class:`onegov.form.fields.UploadField` instance.

    Usage::

        ExpectedFileType('*')  # no check, really
        ExpectedFileType('pdf')  # makes sure the given file is a pdf
    """

    def __init__(self, extensions):
        mimetypes = set(
            types_map.get('.' + ext.lstrip('.'), None) for ext in extensions)
        super().__init__(whitelist=mimetypes)


class ValidFormDefinition(object):
    """ Makes sure the given text is a valid onegov.form definition. """

    message = _("The form could not be parsed.")
    email = _("Define at least one required e-mail field ('E-Mail * = @@@')")
    syntax = _("The syntax on line {line} is not valid.")
    duplicate = _("The field '{label}' exists more than once.")
    reserved = _("'{label}' is a reserved name. Please use a different name.")

    def __init__(self, require_email_field=True, reserved_fields=None):
        self.require_email_field = require_email_field
        self.reserved_fields = reserved_fields or set()

    def __call__(self, form, field):
        if field.data:
            # XXX circular import
            from onegov.form import parse_form

            try:
                form = parse_form(field.data)()
            except InvalidFormSyntax as e:
                field.widget = with_options(
                    field.widget, **{'data-highlight-line': e.line}
                )
                raise ValidationError(
                    field.gettext(self.syntax).format(line=e.line)
                )
            except DuplicateLabelError as e:
                raise ValidationError(
                    field.gettext(self.duplicate).format(label=e.label)
                )
            except AttributeError:
                raise ValidationError(field.gettext(self.message))
            else:
                if self.require_email_field:
                    if not form.has_required_email_field:
                        raise ValidationError(field.gettext(self.email))

                if self.reserved_fields:
                    for formfield_id, formfield in form._fields.items():
                        if formfield_id in self.reserved_fields:

                            raise ValidationError(
                                field.gettext(self.reserved).format(
                                    label=formfield.label.text
                                )
                            )


class StrictOptional(Optional):
    """ A copy of wtform's Optional validator, but with a more strict approach
    to optional validation checking.

    See https://github.com/wtforms/wtforms/issues/350

    """

    def is_missing(self, value):
        if isinstance(value, FieldStorage):
            return False

        if not value:
            return True

        if isinstance(value, string_types):
            return not self.string_check(value)

        return False

    def __call__(self, form, field):
        raw = field.raw_data and field.raw_data[0]
        val = field.data

        # the selectfields have this annyoing habit of coercing all values
        # that are added to them -> this includes the None, which is turned
        # into 'None'
        if isinstance(field, SelectField) and val == 'None':
            val = None

        if self.is_missing(raw) and self.is_missing(val):
            field.errors[:] = []
            raise StopValidation()


class ValidPhoneNumber(object):
    """ Makes sure the given input is valid phone number.

    Expects an :class:`wtforms.StringField` instance.

    """

    message = _("Not a valid phone number.")

    def __init__(self, country='CH'):
        self.country = country

    def __call__(self, form, field):
        if field.data:
            try:
                number = phonenumbers.parse(field.data, self.country)
            except Exception:
                raise ValidationError(self.message)

            valid = (
                phonenumbers.is_valid_number(number)
                and phonenumbers.is_possible_number(number)
            )
            if not valid:
                raise ValidationError(self.message)
